#include <fmt/core.h>
#include <fmt/ranges.h>
#include <stdio.h>

#include <CLI/CLI.hpp>
#include <algorithm>
#include <limits>
#include <optional>

#include "utils.h"

using json = nlohmann::json;
using std::nullopt;
using std::optional;
using std::string;
using std::vector;

auto tables = ExistTables("./tables.json");

template <typename T = int>
class Range {
   public:
    T up;
    T down;

    static T max() { return std::numeric_limits<T>::max(); }
    static T min() { return std::numeric_limits<T>::min(); }

    static Range<T> find_range(const json& j) {
        Range<T> r{.up = std::numeric_limits<T>::min(),
                   .down = std::numeric_limits<T>::max()};
        let type = type_of(j);

        if (type == "小于") {
            r.up = right_value(j);
            r.down = std::numeric_limits<T>::min();
        } else if (type == "大于") {
            r.up = std::numeric_limits<T>::max();
            r.down = right_value(j);
        } else if (type == "相等") {
            r.up = r.down = right_value(j);
        }
        return r;
    }

    Range<T> operator&(const Range<T>& right) const {
        Range<T> new_range{.up = std::min(this->up, right.up),
                           .down = std::max(this->down, right.down)};
        return new_range;
    }
    Range<T> operator|(const Range<T>& right) const {
        Range<T> new_range{.up = std::max(this->up, right.up),
                           .down = std::min(this->down, right.down)};
        return new_range;
    }

    bool empty() const { return this->down > this->up; }
    bool full() const {
        return this->down == std::numeric_limits<T>::min() &&
               this->up == std::numeric_limits<T>::max();
    }
};

json& check_compare_op(json& j) {
    if (j.empty()) {
        return j;
    }
    if (is_column(j)) {
        return j;
    }

    let op_type = op_type_of(j);
    let type = type_of(j);

    auto& left = j["left"];
    auto& right = j["right"];

    left = check_compare_op(left);
    right = check_compare_op(right);
    let left_op_type = op_type_of(left);
    let right_op_type = op_type_of(right);

    if (op_type == OP_LOGIC && left_op_type == OP_COMPARE &&
        right_op_type == OP_COMPARE) {
        if (is_const_column(left["right"]) && is_const_column(right["right"])) {
            // 说明有一个逻辑连接词, 两边都是比较, 并且比较用符号都相同,
            // 并且比较的右侧都是常值
            let left_range = Range<int>::find_range(left);
            let right_range = Range<int>::find_range(right);
            Range<int> new_range;

            if (type == "且") {
                new_range = left_range & right_range;
            } else if (type == "或") {
                new_range = left_range | right_range;
            }
            if (new_range.empty()) {
                j = {{"type", "bool"}, {"value", false}};
            } else if (new_range.full()) {
                j = {{"type", "bool"}, {"value", true}};
            } else {
                if (new_range.max() != new_range.up &&
                    new_range.min() != new_range.down) {
                    j["left"] = {{"type", "小于"},
                                 {"left", j["left"]["left"]},
                                 {"right", new_range.up}};
                    j["right"] = {{"type", "大于"},
                                  {"left", j["right"]["left"]},
                                  {"right", new_range.down}};

                } else if (new_range.max() == new_range.up) {
                    j = {
                        {"type", "大于"},
                        {"left", j["left"]["left"]},
                        {"right", {{"type", "int"}, {"value", new_range.down}}},
                        {"op_type", "bin_cmp_op"}};
                } else if (new_range.min() == new_range.down) {
                    j = {{"type", "小于"},
                         {"left", j["left"]["left"]},
                         {"right", {{"type", "int"}, {"value", new_range.up}}},
                         {"op_type", "bin_cmp_op"}};
                }
            }
        }
    }

    return j;
}

json& check_logic_op(json& j) {
    if (j.empty()) {
        return j;
    }
    let op_type = op_type_of(j);
    if (op_type != OP_LOGIC) {
        return j;
    }
    auto& left = j["left"];
    auto& right = j["right"];

    // 假 或 -> 单独一侧
    // 假 且 -> FALSE
    // 真 或 -> 真
    // 真 且 -> 另一侧
    if (left["type"] == "bool" && left["type"] == "bool") {
        if (j["type"] == "且") {
            j = {{"type", "bool"}, {"value", left["value"] && right["value"]}};
        } else if (j["type"] == "或") {
            j = {{"type", "bool"}, {"value", left["value"] || right["value"]}};
        }
    } else if (left["type"] == "bool" || right["type"] == "bool") {
        if (left["type"] == "bool") {
            let i = right;
            right = left;
            left = i;
        }
        bool right_v = right["value"];

        if (j["type"] == "或" && !right_v) {
            j = left;
        } else if (j["type"] == "且" && !right_v) {
            j = {{"type", "bool"}, {"value", false}};
        } else if (j["type"] == "或" && right_v) {
            j = {{"type", "bool"}, {"value", true}};
        } else if (j["type"] == "且" && right_v) {
            j = left;
        }
    }
    // if (left["type"] == "bool") {
    // }
    return j;
}

bool 谓词下推(json& total_stmt, json& parents_node, json& current_node, json& bro_node);

bool 谓词下推_start(json& total_stmt) {
    return 谓词下推(total_stmt, total_stmt["join_options"]["on"],
                    total_stmt["join_options"]["on"], total_stmt["join_options"]["on"]);
}

// 成功进行谓词下推后返回真
bool 谓词下推(json& total_stmt, json& parents_node, json& current_node, json& bro_node) {
    if (current_node.empty()) {
        return false;
    }
    if (is_column(current_node)) {
        return false;
    }
    // let op_type = op_type_of(current_node);
    // if (op_type != OP_LOGIC) {
    //     return false;
    // }
    auto& left = current_node["left"];
    auto& right = current_node["right"];

    // fmt::print("left: {}\n", left.dump(2));
    // fmt::print("right: {}\n", right.dump(2));

    if (谓词下推(total_stmt, current_node, left, right)) {
        return 谓词下推_start(total_stmt);
    }
    if (谓词下推(total_stmt, current_node, right, left)) {
        return 谓词下推_start(total_stmt);
    }

    let left_col_name = select_column_get_name(left, true);
    let right_col_name = select_column_get_name(right, true);

    const vector<string> all_table_names = total_stmt["table_names"];

    // fmt::print("bro_node: {} parents_node == current_node: {}\n", bro_node.dump(2), parents_node == current_node);

    let left_res = tables.find_col_in_tables(all_table_names, left_col_name);
    let right_res = tables.find_col_in_tables(all_table_names, right_col_name);

    // 两个都在目标表中 -> 可以进行谓词下推
    // 只有一个在目标表中 ＆＆ 另一个是常量 -> 可以进行谓词下推
    if(!left_res.has_value() && !right_res.has_value()) {
        return false;
    }
    if(!left_res.has_value() && !is_const_column(left)) {
        return false;
    }
    if(!right_res.has_value() && !is_const_column(right)) {
        return false;
    }

    if (left_res.has_value()) {
        total_stmt["where"] = current_node;
        if (parents_node == current_node) {
            parents_node = json::object();
        } else {
            parents_node = bro_node;
        }

        return true;
    }
    if (right_res.has_value()) {
        total_stmt["where"] = current_node;
        if (parents_node == current_node) {
            parents_node = json::object();
        } else {
            parents_node = bro_node;
        }
        return true;
    }

    return false;
}
void process_sql(json& j) {
    let type = j.value("type", "none");

    if (type != "select_stmt") {
        throw std::runtime_error(
            fmt::format("Unknown expression type `{}`\n", type));
    }

    j["where"] = check_compare_op(j["where"]);
    j["where"] = check_logic_op(j["where"]);
    if (j.contains("join_options")) {
        谓词下推_start(j);
    }
}

int main(int argc, char** argv) {
    CLI::App app{"sql 优化器"};
    string sql = "", filename = "";

    app.add_option("-f,--file", filename, "输入的 sql 解析后的 json 文件")
        ->check(CLI::ExistingFile);
    app.add_option("-s,--sql", sql, "输入的 sql");

    CLI11_PARSE(app, argc, argv);

    json stmts;
    if (sql.length() > 0) {
        stmts = parse_sql(sql).out;
    } else if (filename.length() > 0) {
        stmts = json::parse(std::ifstream(filename));
    } else {
        // std::cout << "请输入 sql 或者 sql 文件" << std::endl;
        // return 1;
        stmts = parse_sql(
                    "select person.name "
                    "from person join class "
                    "on age=22 and class.id=person.classId;")
                    .out;
    }

    for (auto& stmt : stmts) {
        // for(auto& item : stmt.items()) {
        //     fmt::print("{}: {}\n", item.key(), item.value().dump());
        // }
        process_sql(stmt);
        fmt::print("{}\n", stmt.dump(2));
    }

    return 0;
}
